// Copyright (c) 2017-2021, Lawrence Livermore National Security, LLC and
// other Axom Project Developers. See the top-level LICENSE file for details.
//
// SPDX-License-Identifier: (BSD-3-Clause)

#include "gtest/gtest.h"
#include "axom/slic.hpp"

#include "axom/primal/geometry/Vector.hpp"
#include "axom/primal/geometry/Point.hpp"

#include "axom/core/execution/execution_space.hpp"  // for execution_space traits
#include "axom/core/execution/for_all.hpp"          // for_all()

using namespace axom;

//------------------------------------------------------------------------------
template <typename ExecSpace>
void check_vector_policy()
{
  const int DIM = 3;
  using VectorType = primal::Vector<double, DIM>;

  VectorType* vec =
    axom::allocate<VectorType>(1,
                               axom::execution_space<ExecSpace>::allocatorID());

  axom::for_all<ExecSpace>(
    1,
    AXOM_LAMBDA(int /*i*/) {
      vec[0] = VectorType(-1.0);
      vec[0].negate();
    });

  EXPECT_EQ(vec[0], VectorType(1));
  axom::deallocate(vec);
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_constructors)
{
  static const int DIM = 5;
  typedef double CoordType;
  typedef primal::NumericArray<CoordType, DIM> QArray;
  typedef primal::Vector<CoordType, DIM> QVec;

  QVec vec1;
  EXPECT_EQ(vec1.dimension(), DIM);
  for(int i = 0; i < DIM; ++i)
  {
    EXPECT_EQ(vec1[i], CoordType());
  }

  CoordType val = 5.;
  QVec vec2(val);
  EXPECT_EQ(vec2.dimension(), DIM);
  for(int i = 0; i < DIM; ++i)
  {
    EXPECT_EQ(vec2[i], val);
  }

  CoordType val3 = 5.;
  int halfPt = DIM / 2;
  QVec vec3(val3, halfPt);
  for(int i = 0; i < DIM; ++i)
  {
    CoordType expVal = i < halfPt ? val3 : CoordType();
    EXPECT_EQ(vec3[i], expVal);
  }

  // Compare vector initialized from arbitrary
  CoordType valsArr[DIM] = {12., 23., 34., 45., 56.432};
  QVec arrVec(valsArr);
  for(int i = 0; i < DIM; ++i)
  {
    EXPECT_EQ(arrVec[i], valsArr[i]);
  }

  QVec arrHalfVec(valsArr, halfPt);
  for(int i = 0; i < DIM; ++i)
  {
    CoordType expVal = i < halfPt ? valsArr[i] : CoordType();
    EXPECT_EQ(arrHalfVec[i], expVal);
  }

  // Construct from numeric array
  QArray arr(valsArr);
  QVec vFromNA(arr);
  for(int i = 0; i < DIM; ++i)
  {
    EXPECT_EQ(vFromNA[i], valsArr[i]);
  }

  // Initializer list constructor
  primal::Vector<int, 3> fromInitializerListRightSize = {10, 20, 30};
  for(int i = 0; i < 3; ++i)
  {
    EXPECT_EQ(10 * (i + 1), fromInitializerListRightSize[i]);
  }

  primal::Vector<int, 3> fromInitializerListTooLong = {10, 20, 30, 40};
  for(int i = 0; i < 3; ++i)
  {
    EXPECT_EQ(10 * (i + 1), fromInitializerListTooLong[i]);
  }

  primal::Vector<int, 5> fromInitializerListTooShort = {10, 20};
  for(int i = 0; i < 2; ++i)
  {
    EXPECT_EQ(10 * (i + 1), fromInitializerListTooShort[i]);
  }
  for(int i = 2; i < 5; ++i)
  {
    EXPECT_EQ(0, fromInitializerListTooShort[i]);
  }

  primal::Vector<int, 3> fromInitializerNoEqualsSign {10, 20, 30};
  for(int i = 0; i < 3; ++i)
  {
    EXPECT_EQ(10 * (i + 1), fromInitializerNoEqualsSign[i]);
  }
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_from_points_constructor)
{
  static const int DIM = 5;
  typedef double CoordType;
  typedef primal::Point<CoordType, DIM> QPoint;
  typedef primal::Vector<CoordType, DIM> QVec;

  const double aVal = 12;
  const double bVal = 5;
  QPoint a(aVal);
  QPoint b(bVal);

  QVec vec1(a, b);
  for(int i = 0; i < DIM; ++i)
  {
    EXPECT_EQ(vec1[i], bVal - aVal);
  }

  QVec vec2(b, a);
  for(int i = 0; i < DIM; ++i)
  {
    EXPECT_EQ(vec2[i], aVal - bVal);
  }

  QVec vec3(b);
  QVec vec4(QPoint::zero(), b);
  for(int i = 0; i < DIM; ++i)
  {
    EXPECT_EQ(vec3[i], bVal);
    EXPECT_EQ(vec4[i], bVal);
  }
  EXPECT_EQ(vec3, vec4);
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_normalize)
{
  static const int DIM = 3;
  typedef double CoordType;
  typedef primal::Point<CoordType, DIM> QPoint;
  typedef primal::Vector<CoordType, DIM> QVec;

  EXPECT_DOUBLE_EQ(QVec().unitVector().norm(), 1.0);

  // Find the norm of an arbitrary point
  QPoint p = QPoint::make_point(543.5435, 1566.423532, -432.4);
  QVec vec2(QPoint(), p);
  EXPECT_DOUBLE_EQ(vec2.unitVector().norm(), 1.0);

  // Find the norm of the zero vector
  // Zero vector should become the unit vector when normalized
  QVec vecZero;
  QVec unitVec(1, 1);
  EXPECT_DOUBLE_EQ(vecZero.unitVector().norm(), 1.0);
  EXPECT_EQ(vecZero.unitVector(), unitVec);
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_norm)
{
  static const int DIM = 2;
  typedef double CoordType;
  typedef primal::Point<CoordType, DIM> QPoint;
  typedef primal::Vector<CoordType, DIM> QVec;

  QPoint p1 = QPoint::make_point(3, 0);
  QPoint p2 = QPoint::make_point(0, 4);
  QVec vec(p1, p2);
  EXPECT_DOUBLE_EQ(vec.norm(), 5.0);
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_arithmetic)
{
  static const int DIM = 3;
  typedef double CoordType;
  typedef primal::Point<CoordType, DIM> QPoint;
  typedef primal::Vector<CoordType, DIM> QVec;

  QPoint p1 = QPoint::make_point(3, 0, 1.2);
  QPoint p2 = QPoint::make_point(0, 4, 1.2);
  QPoint pSum = QPoint::make_point(3, 4, 2.4);
  QPoint pDiff = QPoint::make_point(-3, 4, 0);

  CoordType scalar = 5.3;
  QPoint pScalar = QPoint::make_point(scalar * 3, scalar * 4, scalar * 2.4);

  QVec v1(p1);
  QVec v2(p2);

  // testing vector addition and subtraction
  EXPECT_EQ(QVec(p1, p2), v2 - v1);
  EXPECT_EQ(QVec(pDiff), v2 - v1);
  EXPECT_EQ(QVec(pSum), v1 + v2);
  EXPECT_EQ(v1 + v2, v2 + v1);

  QVec vSum(v1);
  vSum += v2;
  EXPECT_EQ(vSum, QVec(pSum));

  QVec vDiff(v2);
  vDiff -= v1;
  EXPECT_EQ(vDiff, QVec(pDiff));

  // Testing scalar multiplication
  EXPECT_EQ(QVec(pScalar), QVec(pSum) * scalar);
  EXPECT_EQ(QVec(pScalar), scalar * QVec(pSum));
  EXPECT_EQ(QVec(pScalar), (v1 + v2) * scalar);

  // Testing unary negation
  EXPECT_EQ(-v1, QVec(p1, QPoint::zero()));
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_inner_product)
{
  static const int DIM = 3;
  typedef double CoordType;
  typedef primal::Point<CoordType, DIM> QPoint;
  typedef primal::Vector<CoordType, DIM> QVec;

  QPoint p1 = QPoint::make_point(3, 0, 1.2);
  QPoint p2 = QPoint::make_point(0, 4, 1.2);

  CoordType expDot = 1.2 * 1.2;

  QVec v1(p1);
  QVec v2(p2);
  EXPECT_EQ(expDot, v1.dot(v2));
  EXPECT_EQ(v2.dot(v1), v1.dot(v2));
  EXPECT_EQ(v1.dot(v2), QVec::dot_product(v1, v2));
  EXPECT_EQ(QVec::dot_product(v1, v2), QVec::dot_product(v2, v1));

  QVec zer;
  EXPECT_EQ(0., zer.dot(zer));
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_outer_product)
{
  static const int DIM = 3;
  typedef double CoordType;
  typedef primal::Point<CoordType, DIM> QPoint;
  typedef primal::Vector<CoordType, DIM> QVec;

  QPoint p1 = QPoint::make_point(3, 0);
  QPoint p2 = QPoint::make_point(0, 4);

  QVec v1(p1);
  QVec v2(p2);

  QVec vRes;
  vRes[2] = 12.;

  EXPECT_EQ(QVec::cross_product(v1, v2), vRes);
  EXPECT_EQ(QVec::cross_product(v2, v1), -vRes);
  EXPECT_EQ(QVec::cross_product(v1, v2), -QVec::cross_product(v2, v1));
}

//------------------------------------------------------------------------------
TEST(primal_vector, vector_zero)
{
  constexpr int DIM = 5;
  typedef primal::Vector<double, DIM> QVec;
  QVec zero {0.0};
  EXPECT_TRUE(zero.is_zero());

  for(int i = 0; i < DIM; ++i)
  {
    QVec notZero = zero;
    notZero[i] = 1e-7;
    EXPECT_FALSE(notZero.is_zero()) << "Wrong when changing index " << i;
  }
}

//------------------------------------------------------------------------------
TEST(primal_vector, add_point_vector)
{
  constexpr int DIM = 3;
  using QPt = primal::Point<double, DIM>;
  using QVec = primal::Vector<double, DIM>;

  {
    QPt one {1.};
    QVec zero {0.0};
    EXPECT_EQ(one, one + zero);
    EXPECT_EQ(one, zero + one);
    EXPECT_EQ(zero + one, one + zero);
  }

  {
    QPt pt {1.23, 4.56, 7.89};
    QVec vec {.23, .56, .89};

    EXPECT_EQ(pt + vec, vec + pt);

    QPt pv = pt + -vec;
    QPt vp = -vec + pt;
    QPt exp {1, 4, 7};
    for(int i = 0; i < DIM; ++i)
    {
      EXPECT_DOUBLE_EQ(pv[i], vp[i]);
      EXPECT_DOUBLE_EQ(exp[i], pv[i]);
    }
  }
}

//------------------------------------------------------------------------------
TEST(primal_vector, subtract_points)
{
  constexpr int DIM = 3;
  using QPt = primal::Point<double, DIM>;
  using QVec = primal::Vector<double, DIM>;

  {
    QPt zeroPt {0.0};
    QPt onePt {1.0};

    QVec oneVec {1.0};

    EXPECT_EQ(oneVec, onePt - zeroPt);
    EXPECT_EQ(-oneVec, zeroPt - onePt);
  }

  {
    QPt head {1.23, 4.56, 7.89};
    QPt tail {.23, .56, .89};

    QVec diff = head - tail;
    QVec exp {1, 4, 7};
    QVec ctor(tail, head);

    for(int i = 0; i < DIM; ++i)
    {
      EXPECT_DOUBLE_EQ(exp[i], diff[i]);
      EXPECT_DOUBLE_EQ(ctor[i], diff[i]);
    }
  }
}

//------------------------------------------------------------------------------
AXOM_CUDA_TEST(primal_numeric_array, numeric_array_check_policies)
{
  using seq_exec = axom::SEQ_EXEC;
  check_vector_policy<seq_exec>();

#if defined(AXOM_USE_RAJA) && defined(AXOM_USE_OPENMP) && \
  defined(RAJA_ENABLE_OPENMP)

  using omp_exec = axom::OMP_EXEC;
  check_vector_policy<omp_exec>();

#endif

#if defined(AXOM_USE_RAJA) && defined(AXOM_USE_CUDA) && \
  defined(RAJA_ENABLE_CUDA) && defined(AXOM_USE_UMPIRE)

  using cuda_exec = axom::CUDA_EXEC<512>;

  check_vector_policy<cuda_exec>();
#endif
}

//----------------------------------------------------------------------

int main(int argc, char* argv[])
{
  int result = 0;

  ::testing::InitGoogleTest(&argc, argv);
  axom::slic::SimpleLogger logger;

  result = RUN_ALL_TESTS();

  return result;
}
